@@ -1,6 +1,6 @@
 # tags.py - read tag info from local repository
 #
-# Copyright 2009 Matt Mackall <mpm@selenic.com>
+# Copyright 2009 Olivia Mackall <olivia@selenic.com>
 # Copyright 2009 Greg Ward <greg@gerg.ca>
 #
 # This software may be used and distributed according to the terms of the
@@ -10,15 +10,13 @@
 # Eventually, it could take care of updating (adding/removing/moving)
 # tags too.
 
-from __future__ import absolute_import
 
-import errno
+import binascii
 import io
 
 from .node import (
     bin,
     hex,
-    nullid,
     nullrev,
     short,
 )
@@ -30,9 +28,7 @@ from . import (
     scmutil,
     util,
 )
-from .utils import (
-    stringutil,
-)
+from .utils import stringutil
 
 # Tags computation can be expensive and caches exist to make it fast in
 # the common case.
@@ -83,6 +79,7 @@ from .utils import (
 # The most recent changeset (in terms of revlog ordering for the head
 # setting it) for each tag is last.
 
+
 def fnoderevs(ui, repo, revs):
     """return the list of '.hgtags' fnodes used in a set revisions
 
@@ -95,15 +92,17 @@ def fnoderevs(ui, repo, revs):
     fnodes = _filterfnodes(fnodes, nodes)
     return fnodes
 
-def _nulltonone(value):
+
+def _nulltonone(repo, value):
     """convert nullid to None
 
     For tag value, nullid means "deleted". This small utility function helps
     translating that to None."""
-    if value == nullid:
+    if value == repo.nullid:
         return None
     return value
 
+
 def difftags(ui, repo, oldfnodes, newfnodes):
     """list differences between tags expressed in two set of file-nodes
 
@@ -121,19 +120,20 @@ def difftags(ui, repo, oldfnodes, newfnodes):
     # list of (tag, old, new): None means missing
     entries = []
     for tag, (new, __) in newtags.items():
-        new = _nulltonone(new)
+        new = _nulltonone(repo, new)
         old, __ = oldtags.pop(tag, (None, None))
-        old = _nulltonone(old)
+        old = _nulltonone(repo, old)
         if old != new:
             entries.append((tag, old, new))
     # handle deleted tags
     for tag, (old, __) in oldtags.items():
-        old = _nulltonone(old)
+        old = _nulltonone(repo, old)
         if old is not None:
             entries.append((tag, old, None))
     entries.sort()
     return entries
 
+
 def writediff(fp, difflist):
     """write tags diff information to a file.
 
@@ -153,10 +153,10 @@ def writediff(fp, difflist):
 
     See documentation of difftags output for details about the input.
     """
-    add = '+A %s %s\n'
-    remove = '-R %s %s\n'
-    updateold = '-M %s %s\n'
-    updatenew = '+M %s %s\n'
+    add = b'+A %s %s\n'
+    remove = b'-R %s %s\n'
+    updateold = b'-M %s %s\n'
+    updatenew = b'+M %s %s\n'
     for tag, old, new in difflist:
         # translate to hex
         if old is not None:
@@ -172,13 +172,14 @@ def writediff(fp, difflist):
             fp.write(updateold % (old, tag))
             fp.write(updatenew % (new, tag))
 
+
 def findglobaltags(ui, repo):
-    '''Find global tags in a repo: return a tagsmap
+    """Find global tags in a repo: return a tagsmap
 
     tagsmap: tag name to (node, hist) 2-tuples.
 
     The tags cache is read and updated as a side-effect of calling.
-    '''
+    """
     (heads, tagfnode, valid, cachetags, shouldwrite) = _readtagcache(ui, repo)
     if cachetags is not None:
         assert not shouldwrite
@@ -190,8 +191,9 @@ def findglobaltags(ui, repo):
         return alltags
 
     for head in reversed(heads):  # oldest to newest
-        assert head in repo.changelog.nodemap, (
-               "tag cache returned bogus head %s" % short(head))
+        assert repo.changelog.index.has_node(
+            head
+        ), b"tag cache returned bogus head %s" % short(head)
     fnodes = _filterfnodes(tagfnode, reversed(heads))
     alltags = _tagsfromfnodes(ui, repo, fnodes)
 
@@ -200,6 +202,7 @@ def findglobaltags(ui, repo):
         _writetagcache(ui, repo, valid, alltags)
     return alltags
 
+
 def _filterfnodes(tagfnode, nodes):
     """return a list of unique fnodes
 
@@ -215,6 +218,7 @@ def _filterfnodes(tagfnode, nodes):
             fnodes.append(fnode)
     return fnodes
 
+
 def _tagsfromfnodes(ui, repo, fnodes):
     """return a tagsmap from a list of file-node
 
@@ -225,27 +229,26 @@ def _tagsfromfnodes(ui, repo, fnodes):
     fctx = None
     for fnode in fnodes:
         if fctx is None:
-            fctx = repo.filectx('.hgtags', fileid=fnode)
+            fctx = repo.filectx(b'.hgtags', fileid=fnode)
         else:
             fctx = fctx.filectx(fnode)
         filetags = _readtags(ui, repo, fctx.data().splitlines(), fctx)
         _updatetags(filetags, alltags)
     return alltags
 
+
 def readlocaltags(ui, repo, alltags, tagtypes):
     '''Read local tags in repo. Update alltags and tagtypes.'''
     try:
-        data = repo.vfs.read("localtags")
-    except IOError as inst:
-        if inst.errno != errno.ENOENT:
-            raise
+        data = repo.vfs.read(b"localtags")
+    except FileNotFoundError:
         return
 
     # localtags is in the local encoding; re-encode to UTF-8 on
     # input for consistency with the rest of this module.
     filetags = _readtags(
-        ui, repo, data.splitlines(), "localtags",
-        recode=encoding.fromlocal)
+        ui, repo, data.splitlines(), b"localtags", recode=encoding.fromlocal
+    )
 
     # remove tags pointing to invalid nodes
     cl = repo.changelog
@@ -255,10 +258,11 @@ def readlocaltags(ui, repo, alltags, tagtypes):
         except (LookupError, ValueError):
             del filetags[t]
 
-    _updatetags(filetags, alltags, 'local', tagtypes)
+    _updatetags(filetags, alltags, b'local', tagtypes)
+
 
 def _readtaghist(ui, repo, lines, fn, recode=None, calcnodelines=False):
-    '''Read tag definitions from a file (or any source of lines).
+    """Read tag definitions from a file (or any source of lines).
 
     This function returns two sortdicts with similar information:
 
@@ -274,31 +278,31 @@ def _readtaghist(ui, repo, lines, fn, recode=None, calcnodelines=False):
     When calcnodelines is False the hextaglines dict is not calculated (an
     empty dict is returned). This is done to improve this function's
     performance in cases where the line numbers are not needed.
-    '''
+    """
 
     bintaghist = util.sortdict()
     hextaglines = util.sortdict()
     count = 0
 
     def dbg(msg):
-        ui.debug("%s, line %d: %s\n" % (fn, count, msg))
+        ui.debug(b"%s, line %d: %s\n" % (fn, count, msg))
 
     for nline, line in enumerate(lines):
         count += 1
         if not line:
             continue
         try:
-            (nodehex, name) = line.split(" ", 1)
+            (nodehex, name) = line.split(b" ", 1)
         except ValueError:
-            dbg("cannot parse entry")
+            dbg(b"cannot parse entry")
             continue
         name = name.strip()
         if recode:
             name = recode(name)
         try:
             nodebin = bin(nodehex)
-        except TypeError:
-            dbg("node '%s' is not well formed" % nodehex)
+        except binascii.Error:
+            dbg(b"node '%s' is not well formed" % nodehex)
             continue
 
         # update filetags
@@ -314,17 +318,19 @@ def _readtaghist(ui, repo, lines, fn, recode=None, calcnodelines=False):
         bintaghist[name].append(nodebin)
     return bintaghist, hextaglines
 
+
 def _readtags(ui, repo, lines, fn, recode=None, calcnodelines=False):
-    '''Read tag definitions from a file (or any source of lines).
+    """Read tag definitions from a file (or any source of lines).
 
     Returns a mapping from tag name to (node, hist).
 
     "node" is the node id from the last line read for that name. "hist"
     is the list of node ids previously associated with it (in file order).
     All node ids are binary, not hex.
-    '''
-    filetags, nodelines = _readtaghist(ui, repo, lines, fn, recode=recode,
-                                       calcnodelines=calcnodelines)
+    """
+    filetags, nodelines = _readtaghist(
+        ui, repo, lines, fn, recode=recode, calcnodelines=calcnodelines
+    )
     # util.sortdict().__setitem__ is much slower at replacing then inserting
     # new entries. The difference can matter if there are thousands of tags.
     # Create a new sortdict to avoid the performance penalty.
@@ -333,6 +339,7 @@ def _readtags(ui, repo, lines, fn, recode=None, calcnodelines=False):
         newtags[tag] = (taghist[-1], taghist[:-1])
     return newtags
 
+
 def _updatetags(filetags, alltags, tagtype=None, tagtypes=None):
     """Incorporate the tag info read from one file into dictionnaries
 
@@ -344,7 +351,7 @@ def _updatetags(filetags, alltags, tagtype=None, tagtypes=None):
     if tagtype is None:
         assert tagtypes is None
 
-    for name, nodehist in filetags.iteritems():
+    for name, nodehist in filetags.items():
         if name not in alltags:
             alltags[name] = nodehist
             if tagtype is not None:
@@ -357,23 +364,28 @@ def _updatetags(filetags, alltags, tagtype=None, tagtypes=None):
         # otherwise we win because we're tip-most
         anode, ahist = nodehist
         bnode, bhist = alltags[name]
-        if (bnode != anode and anode in bhist and
-            (bnode not in ahist or len(bhist) > len(ahist))):
+        if (
+            bnode != anode
+            and anode in bhist
+            and (bnode not in ahist or len(bhist) > len(ahist))
+        ):
             anode = bnode
         elif tagtype is not None:
             tagtypes[name] = tagtype
         ahist.extend([n for n in bhist if n not in ahist])
         alltags[name] = anode, ahist
 
+
 def _filename(repo):
     """name of a tagcache file for a given repo or repoview"""
-    filename = 'tags2'
+    filename = b'tags2'
     if repo.filtername:
-        filename = '%s-%s' % (filename, repo.filtername)
+        filename = b'%s-%s' % (filename, repo.filtername)
     return filename
 
+
 def _readtagcache(ui, repo):
-    '''Read the tag cache.
+    """Read the tag cache.
 
     Returns a tuple (heads, fnodes, validinfo, cachetags, shouldwrite).
 
@@ -389,9 +401,9 @@ def _readtagcache(ui, repo):
 
     If the cache is not up to date, the caller is responsible for reading tag
     info from each returned head. (See findglobaltags().)
-    '''
+    """
     try:
-        cachefile = repo.cachevfs(_filename(repo), 'r')
+        cachefile = repo.cachevfs(_filename(repo), b'r')
         # force reading the file for static-http
         cachelines = iter(cachefile)
     except IOError:
@@ -419,21 +431,23 @@ def _readtagcache(ui, repo):
     # (Unchanged tip trivially means no changesets have been added.
     # But, thanks to localrepository.destroyed(), it also means none
     # have been destroyed by strip or rollback.)
-    if (cacherev == tiprev
-            and cachenode == tipnode
-            and cachehash == scmutil.filteredhash(repo, tiprev)):
+    if (
+        cacherev == tiprev
+        and cachenode == tipnode
+        and cachehash == scmutil.filteredhash(repo, tiprev)
+    ):
         tags = _readtags(ui, repo, cachelines, cachefile.name)
         cachefile.close()
         return (None, None, None, tags, False)
     if cachefile:
-        cachefile.close()               # ignore rest of file
+        cachefile.close()  # ignore rest of file
 
     valid = (tiprev, tipnode, scmutil.filteredhash(repo, tiprev))
 
     repoheads = repo.heads()
     # Case 2 (uncommon): empty repo; get out quickly and don't bother
     # writing an empty cache.
-    if repoheads == [nullid]:
+    if repoheads == [repo.nullid]:
         return ([], {}, valid, {}, False)
 
     # Case 3 (uncommon): cache file missing or empty.
@@ -449,12 +463,11 @@ def _readtagcache(ui, repo):
 
     # N.B. in case 4 (nodes destroyed), "new head" really means "newly
     # exposed".
-    if not len(repo.file('.hgtags')):
+    if not len(repo.file(b'.hgtags')):
         # No tags have ever been committed, so we can avoid a
         # potentially expensive search.
         return ([], {}, valid, None, True)
 
-
     # Now we have to lookup the .hgtags filenode for every new head.
     # This is the most expensive part of finding tags, so performance
     # depends primarily on the size of newheads.  Worst case: no cache
@@ -466,6 +479,7 @@ def _readtagcache(ui, repo):
     # cachefnode to get to each .hgtags revision quickly.
     return (repoheads, cachefnode, valid, None, True)
 
+
 def _getfnodes(ui, repo, nodes):
     """return .hgtags fnodes for a list of changeset nodes
 
@@ -475,50 +489,79 @@ def _getfnodes(ui, repo, nodes):
     starttime = util.timer()
     fnodescache = hgtagsfnodescache(repo.unfiltered())
     cachefnode = {}
+    validated_fnodes = set()
+    unknown_entries = set()
+
+    flog = None
     for node in nodes:
         fnode = fnodescache.getfnode(node)
-        if fnode != nullid:
+        if fnode != repo.nullid:
+            if fnode not in validated_fnodes:
+                if flog is None:
+                    flog = repo.file(b'.hgtags')
+                if flog.hasnode(fnode):
+                    validated_fnodes.add(fnode)
+                else:
+                    unknown_entries.add(node)
             cachefnode[node] = fnode
 
+    if unknown_entries:
+        fixed_nodemap = fnodescache.refresh_invalid_nodes(unknown_entries)
+        for node, fnode in fixed_nodemap.items():
+            if fnode != repo.nullid:
+                cachefnode[node] = fnode
+
     fnodescache.write()
 
     duration = util.timer() - starttime
-    ui.log('tagscache',
-           '%d/%d cache hits/lookups in %0.4f seconds\n',
-           fnodescache.hitcount, fnodescache.lookupcount, duration)
+    ui.log(
+        b'tagscache',
+        b'%d/%d cache hits/lookups in %0.4f seconds\n',
+        fnodescache.hitcount,
+        fnodescache.lookupcount,
+        duration,
+    )
     return cachefnode
 
+
 def _writetagcache(ui, repo, valid, cachetags):
     filename = _filename(repo)
     try:
-        cachefile = repo.cachevfs(filename, 'w', atomictemp=True)
+        cachefile = repo.cachevfs(filename, b'w', atomictemp=True)
     except (OSError, IOError):
         return
 
-    ui.log('tagscache', 'writing .hg/cache/%s with %d tags\n',
-           filename, len(cachetags))
+    ui.log(
+        b'tagscache',
+        b'writing .hg/cache/%s with %d tags\n',
+        filename,
+        len(cachetags),
+    )
 
     if valid[2]:
-        cachefile.write('%d %s %s\n' % (valid[0], hex(valid[1]), hex(valid[2])))
+        cachefile.write(
+            b'%d %s %s\n' % (valid[0], hex(valid[1]), hex(valid[2]))
+        )
     else:
-        cachefile.write('%d %s\n' % (valid[0], hex(valid[1])))
+        cachefile.write(b'%d %s\n' % (valid[0], hex(valid[1])))
 
     # Tag names in the cache are in UTF-8 -- which is the whole reason
     # we keep them in UTF-8 throughout this module.  If we converted
     # them local encoding on input, we would lose info writing them to
     # the cache.
-    for (name, (node, hist)) in sorted(cachetags.iteritems()):
+    for (name, (node, hist)) in sorted(cachetags.items()):
         for n in hist:
-            cachefile.write("%s %s\n" % (hex(n), name))
-        cachefile.write("%s %s\n" % (hex(node), name))
+            cachefile.write(b"%s %s\n" % (hex(n), name))
+        cachefile.write(b"%s %s\n" % (hex(node), name))
 
     try:
         cachefile.close()
     except (OSError, IOError):
         pass
 
+
 def tag(repo, names, node, message, local, user, date, editor=False):
-    '''tag a revision with one or more symbolic names.
+    """tag a revision with one or more symbolic names.
 
     names is a list of strings or, when adding a single tag, names may be a
     string.
@@ -536,70 +579,81 @@ def tag(repo, names, node, message, local, user, date, editor=False):
 
     user: name of user to use if committing
 
-    date: date tuple to use if committing'''
+    date: date tuple to use if committing"""
 
     if not local:
-        m = matchmod.exact(['.hgtags'])
-        if any(repo.status(match=m, unknown=True, ignored=True)):
-            raise error.Abort(_('working copy of .hgtags is changed'),
-                             hint=_('please commit .hgtags manually'))
+        m = matchmod.exact([b'.hgtags'])
+        st = repo.status(match=m, unknown=True, ignored=True)
+        if any(
+            (
+                st.modified,
+                st.added,
+                st.removed,
+                st.deleted,
+                st.unknown,
+                st.ignored,
+            )
+        ):
+            raise error.Abort(
+                _(b'working copy of .hgtags is changed'),
+                hint=_(b'please commit .hgtags manually'),
+            )
 
     with repo.wlock():
-        repo.tags() # instantiate the cache
-        _tag(repo, names, node, message, local, user, date,
-             editor=editor)
+        repo.tags()  # instantiate the cache
+        _tag(repo, names, node, message, local, user, date, editor=editor)
 
-def _tag(repo, names, node, message, local, user, date, extra=None,
-         editor=False):
+
+def _tag(
+    repo, names, node, message, local, user, date, extra=None, editor=False
+):
     if isinstance(names, bytes):
         names = (names,)
 
     branches = repo.branchmap()
     for name in names:
-        repo.hook('pretag', throw=True, node=hex(node), tag=name,
-                  local=local)
+        repo.hook(b'pretag', throw=True, node=hex(node), tag=name, local=local)
         if name in branches:
-            repo.ui.warn(_("warning: tag %s conflicts with existing"
-            " branch name\n") % name)
+            repo.ui.warn(
+                _(b"warning: tag %s conflicts with existing branch name\n")
+                % name
+            )
 
     def writetags(fp, names, munge, prevtags):
         fp.seek(0, io.SEEK_END)
-        if prevtags and not prevtags.endswith('\n'):
-            fp.write('\n')
+        if prevtags and not prevtags.endswith(b'\n'):
+            fp.write(b'\n')
         for name in names:
             if munge:
                 m = munge(name)
             else:
                 m = name
 
-            if (repo._tagscache.tagtypes and
-                name in repo._tagscache.tagtypes):
-                old = repo.tags().get(name, nullid)
-                fp.write('%s %s\n' % (hex(old), m))
-            fp.write('%s %s\n' % (hex(node), m))
+            if repo._tagscache.tagtypes and name in repo._tagscache.tagtypes:
+                old = repo.tags().get(name, repo.nullid)
+                fp.write(b'%s %s\n' % (hex(old), m))
+            fp.write(b'%s %s\n' % (hex(node), m))
         fp.close()
 
-    prevtags = ''
+    prevtags = b''
     if local:
         try:
-            fp = repo.vfs('localtags', 'r+')
+            fp = repo.vfs(b'localtags', b'r+')
         except IOError:
-            fp = repo.vfs('localtags', 'a')
+            fp = repo.vfs(b'localtags', b'a')
         else:
             prevtags = fp.read()
 
         # local tags are stored in the current charset
         writetags(fp, names, None, prevtags)
         for name in names:
-            repo.hook('tag', node=hex(node), tag=name, local=local)
+            repo.hook(b'tag', node=hex(node), tag=name, local=local)
         return
 
     try:
-        fp = repo.wvfs('.hgtags', 'rb+')
-    except IOError as e:
-        if e.errno != errno.ENOENT:
-            raise
-        fp = repo.wvfs('.hgtags', 'ab')
+        fp = repo.wvfs(b'.hgtags', b'rb+')
+    except FileNotFoundError:
+        fp = repo.wvfs(b'.hgtags', b'ab')
     else:
         prevtags = fp.read()
 
@@ -610,23 +664,26 @@ def _tag(repo, names, node, message, local, user, date, extra=None,
 
     repo.invalidatecaches()
 
-    if '.hgtags' not in repo.dirstate:
-        repo[None].add(['.hgtags'])
+    if b'.hgtags' not in repo.dirstate:
+        repo[None].add([b'.hgtags'])
 
-    m = matchmod.exact(['.hgtags'])
-    tagnode = repo.commit(message, user, date, extra=extra, match=m,
-                          editor=editor)
+    m = matchmod.exact([b'.hgtags'])
+    tagnode = repo.commit(
+        message, user, date, extra=extra, match=m, editor=editor
+    )
 
     for name in names:
-        repo.hook('tag', node=hex(node), tag=name, local=local)
+        repo.hook(b'tag', node=hex(node), tag=name, local=local)
 
     return tagnode
 
-_fnodescachefile = 'hgtagsfnodes1'
-_fnodesrecsize = 4 + 20 # changeset fragment + filenode
-_fnodesmissingrec = '\xff' * 24
 
-class hgtagsfnodescache(object):
+_fnodescachefile = b'hgtagsfnodes1'
+_fnodesrecsize = 4 + 20  # changeset fragment + filenode
+_fnodesmissingrec = b'\xff' * 24
+
+
+class hgtagsfnodescache:
     """Persistent cache mapping revisions to .hgtags filenodes.
 
     The cache is an array of records. Each item in the array corresponds to
@@ -645,6 +702,7 @@ class hgtagsfnodescache(object):
     Instances behave like lists. ``c[i]`` works where i is a rev or
     changeset node. Missing indexes are populated automatically on access.
     """
+
     def __init__(self, repo):
         assert repo.filtername is None
 
@@ -654,11 +712,10 @@ class hgtagsfnodescache(object):
         self.lookupcount = 0
         self.hitcount = 0
 
-
         try:
             data = repo.cachevfs.read(_fnodescachefile)
         except (OSError, IOError):
-            data = ""
+            data = b""
         self._raw = bytearray(data)
 
         # The end state of self._raw is an array that is of the exact length
@@ -673,41 +730,47 @@ class hgtagsfnodescache(object):
 
         self._dirtyoffset = None
 
-        if rawlen < wantedlen:
-            self._dirtyoffset = rawlen
-            self._raw.extend('\xff' * (wantedlen - rawlen))
-        elif rawlen > wantedlen:
+        rawlentokeep = min(
+            wantedlen, (rawlen // _fnodesrecsize) * _fnodesrecsize
+        )
+        if rawlen > rawlentokeep:
             # There's no easy way to truncate array instances. This seems
             # slightly less evil than copying a potentially large array slice.
-            for i in range(rawlen - wantedlen):
+            for i in range(rawlen - rawlentokeep):
                 self._raw.pop()
-            self._dirtyoffset = len(self._raw)
+            rawlen = len(self._raw)
+            self._dirtyoffset = rawlen
+        if rawlen < wantedlen:
+            if self._dirtyoffset is None:
+                self._dirtyoffset = rawlen
+            # TODO: zero fill entire record, because it's invalid not missing?
+            self._raw.extend(b'\xff' * (wantedlen - rawlen))
 
     def getfnode(self, node, computemissing=True):
         """Obtain the filenode of the .hgtags file at a specified revision.
 
         If the value is in the cache, the entry will be validated and returned.
         Otherwise, the filenode will be computed and returned unless
-        "computemissing" is False, in which case None will be returned without
+        "computemissing" is False.  In that case, None will be returned if
+        the entry is missing or False if the entry is invalid without
         any potentially expensive computation being performed.
 
         If an .hgtags does not exist at the specified revision, nullid is
         returned.
         """
-        if node == nullid:
-            return nullid
+        if node == self._repo.nullid:
+            return node
 
-        ctx = self._repo[node]
-        rev = ctx.rev()
+        rev = self._repo.changelog.rev(node)
 
         self.lookupcount += 1
 
         offset = rev * _fnodesrecsize
-        record = '%s' % self._raw[offset:offset + _fnodesrecsize]
+        record = b'%s' % self._raw[offset : offset + _fnodesrecsize]
         properprefix = node[0:4]
 
         # Validate and return existing entry.
-        if record != _fnodesmissingrec:
+        if record != _fnodesmissingrec and len(record) == _fnodesrecsize:
             fileprefix = record[0:4]
 
             if fileprefix == properprefix:
@@ -719,8 +782,19 @@ class hgtagsfnodescache(object):
         # If we get here, the entry is either missing or invalid.
 
         if not computemissing:
+            if record != _fnodesmissingrec:
+                return False
             return None
 
+        fnode = self._computefnode(node)
+        self._writeentry(offset, properprefix, fnode)
+        return fnode
+
+    def _computefnode(self, node):
+        """Finds the tag filenode for a node which is missing or invalid
+        in cache"""
+        ctx = self._repo[node]
+        rev = ctx.rev()
         fnode = None
         cl = self._repo.changelog
         p1rev, p2rev = cl._uncheckedparentrevs(rev)
@@ -730,26 +804,24 @@ class hgtagsfnodescache(object):
             # There is some no-merge changeset where p1 is null and p2 is set
             # Processing them as merge is just slower, but still gives a good
             # result.
-            p2node = cl.node(p1rev)
+            p2node = cl.node(p2rev)
             p2fnode = self.getfnode(p2node, computemissing=False)
             if p1fnode != p2fnode:
                 # we cannot rely on readfast because we don't know against what
                 # parent the readfast delta is computed
                 p1fnode = None
-        if p1fnode is not None:
+        if p1fnode:
             mctx = ctx.manifestctx()
-            fnode = mctx.readfast().get('.hgtags')
+            fnode = mctx.readfast().get(b'.hgtags')
             if fnode is None:
                 fnode = p1fnode
         if fnode is None:
             # Populate missing entry.
             try:
-                fnode = ctx.filenode('.hgtags')
+                fnode = ctx.filenode(b'.hgtags')
             except error.LookupError:
                 # No .hgtags file on this revision.
-                fnode = nullid
-
-        self._writeentry(offset, properprefix, fnode)
+                fnode = self._repo.nullid
         return fnode
 
     def setfnode(self, node, fnode):
@@ -763,10 +835,25 @@ class hgtagsfnodescache(object):
 
         self._writeentry(ctx.rev() * _fnodesrecsize, node[0:4], fnode)
 
+    def refresh_invalid_nodes(self, nodes):
+        """recomputes file nodes for a given set of nodes which has unknown
+        filenodes for them in the cache
+        Also updates the in-memory cache with the correct filenode.
+        Caller needs to take care about calling `.write()` so that updates are
+        persisted.
+        Returns a map {node: recomputed fnode}
+        """
+        fixed_nodemap = {}
+        for node in nodes:
+            fnode = self._computefnode(node)
+            fixed_nodemap[node] = fnode
+            self.setfnode(node, fnode)
+        return fixed_nodemap
+
     def _writeentry(self, offset, prefix, fnode):
         # Slices on array instances only accept other array.
         entry = bytearray(prefix + fnode)
-        self._raw[offset:offset + _fnodesrecsize] = entry
+        self._raw[offset : offset + _fnodesrecsize] = entry
         # self._dirtyoffset could be None.
         self._dirtyoffset = min(self._dirtyoffset or 0, offset or 0)
 
@@ -779,39 +866,46 @@ class hgtagsfnodescache(object):
         if self._dirtyoffset is None:
             return
 
-        data = self._raw[self._dirtyoffset:]
+        data = self._raw[self._dirtyoffset :]
         if not data:
             return
 
         repo = self._repo
 
         try:
-            lock = repo.wlock(wait=False)
+            lock = repo.lock(wait=False)
         except error.LockError:
-            repo.ui.log('tagscache', 'not writing .hg/cache/%s because '
-                        'lock cannot be acquired\n' % (_fnodescachefile))
+            repo.ui.log(
+                b'tagscache',
+                b'not writing .hg/cache/%s because '
+                b'lock cannot be acquired\n' % _fnodescachefile,
+            )
             return
 
         try:
-            f = repo.cachevfs.open(_fnodescachefile, 'ab')
+            f = repo.cachevfs.open(_fnodescachefile, b'ab')
             try:
                 # if the file has been truncated
                 actualoffset = f.tell()
                 if actualoffset < self._dirtyoffset:
                     self._dirtyoffset = actualoffset
-                    data = self._raw[self._dirtyoffset:]
+                    data = self._raw[self._dirtyoffset :]
                 f.seek(self._dirtyoffset)
                 f.truncate()
-                repo.ui.log('tagscache',
-                            'writing %d bytes to cache/%s\n' % (
-                            len(data), _fnodescachefile))
+                repo.ui.log(
+                    b'tagscache',
+                    b'writing %d bytes to cache/%s\n'
+                    % (len(data), _fnodescachefile),
+                )
                 f.write(data)
                 self._dirtyoffset = None
             finally:
                 f.close()
         except (IOError, OSError) as inst:
-            repo.ui.log('tagscache',
-                        "couldn't write cache/%s: %s\n" % (
-                            _fnodescachefile, stringutil.forcebytestr(inst)))
+            repo.ui.log(
+                b'tagscache',
+                b"couldn't write cache/%s: %s\n"
+                % (_fnodescachefile, stringutil.forcebytestr(inst)),
+            )
         finally:
             lock.release()
